(* File:     Typechecker.sml
   Author:   Nicholas Wolverson
   Date:     March-June 2008. Subsequently tweaked by JRL.
   Prototype typechecker for Lingay, including argument safety and
   new-safety checking. *)

structure Typechecker = struct

infix <:
infix </:
infix =:=
infix <:>

local open Types; open Absyn in

val debug = ref false
fun debugPrint s = if !debug then TextIO.print s else ()

val e1Debug = ref (NONE : (errpos*ty*bool) exp option) ;
val e2Debug = ref (NONE : (errpos*ty*bool) exp option) ;

datatype st_tag = ST_none | ST_safe | ST_argsafe | ST_writesafe

fun fst (a,b) = a and snd (a,b) = b and id x = x
fun tryDelete (set, x) = VarSet.delete(set, x) handle NotFound => set

(* Type environments. Simply a type name => type mapping *)

val empty_tenv = []
fun extendTy env x t = (x,t)::env
fun extendTyList env newenv = List.revAppend(newenv,env)
fun lookupTy (env:(var * ty) list) (x:var) = 
    (case List.find (fn (y,_) => y=x) env of SOME (_,t) => SOME t
					   | NONE => NONE)
fun lookupTyE pos env x = 
    case lookupTy env x of SOME t => t 
			 | NONE => raise Untypeable (pos, "Unknown type name: "^x)


(* Environments. Should really check status of type on insertion and maintain rather than recompute. *)

datatype var_safe = 
	 Unsafevar | Safevar | Writevar | RWvar | Argsafevar
datatype var_newsafe = 
	 Newunsafevar | Newsafevar | Halfnewsafevar | Weaknewsafevar
	 (* Unsafe = no tag *) 

type var_bisafe = var_safe * var_newsafe

fun tagtostring Ssyn.None = ""
  | tagtostring Ssyn.Safe = "safe"
  | tagtostring Ssyn.Writesafe = "writesafe"
  | tagtostring Ssyn.Argsafe = "argsafe"
  | tagtostring Ssyn.Metasafe = "metasafe"
  | tagtostring Ssyn.MetasafeLin = "metasafe (linear)"
fun vtagtostring Unsafevar = "(none)"
  | vtagtostring Safevar = "safe"
  | vtagtostring Argsafevar = "argsafe"
  | vtagtostring Writevar = "writevar"
  | vtagtostring Metasafevar = "metasafe"

val empty_env = []
fun extend env x t tag = (x,t,tag)::env
fun extendList env newenv = List.revAppend(newenv,env)
fun lookup env x = 
    (case List.find (fn (y,_,_) => y=x) env of SOME (_,t,tag) => SOME (t,tag)
                         | NONE => NONE)

val forest_rep = FunTy(IntTy, ProdTy([ ("label", IntTy), ("rest",  ForestTy)], []))
fun is_forest_rep (FunTy(IntTy, ProdTy(tys,[]))) =
    (case prod_match tys [("label", (fn t=>SOME (t=IntTy))),
			 ("rest", (fn t=>SOME (t=ForestTy)))]
      of SOME l => List.all id l
       | NONE => false)
  | is_forest_rep _ = false

val unitty = ProdTy([],[])

fun is_value (Var _) = true
  | is_value (Const _) = true
  | is_value (Product(_,es)) = List.all (is_value o snd) es
  | is_value (Inject(_,_,e)) = is_value e
  | is_value (Prom(_,e)) = is_value e
  | is_value (Fold(_,_,e)) = is_value e
  | is_value (Fun(_,_,_,e)) = true
  | is_value (Cast(_,e,_)) = is_value e
  | is_value (Constr(_,e)) = is_value e
  | is_value (Triv _) = true
  | is_value (Extend(_,_,Triv _, e1, e2)) = is_value e1 andalso is_value e2
  | is_value (App(_,e,_)) = 
             (case e of Der(_,Var(_,"MAKE_METASAFE")) => true | _ => false)
  | is_value _ = false

fun check_value pos e = if is_value e then () 
			else raise Untypeable(pos, "Should be a value: "^(exptostring e))
						       
infix --
fun extend env x t = (x,t)::env
fun extendList env newenv = List.revAppend(map (fn (x,t) => (x,t)) newenv,env) 
fun lookup env x = 
    (case List.find (fn (y,_) => y=x) env of SOME (_,t) => SOME (t)
                                           | NONE => NONE)
fun alevel env ee = 
    let 
        fun n -- m = if n - m < 0 then 0 else n - m 
        val level = alevel env
        val max = foldl (fn (x,m) => if x > m then x else m) 0 
        val alevels = map (alevel env)
        val maxlevels = max o alevels
    in
        case ee of Var(_,x) => (case lookup env x of SOME n => n 
                                                   | NONE => raise Impossible ("alevel: Unknown variable "^x))
                 | Const _ => 0
                 | Product(_,es) => maxlevels (map snd es)
                 | Merge (_,e1,e2) => maxlevels [e1,e2]
                 | Split (_,e,binds,rest,e') => 
                   let val elevel = alevel env e
                   in 
                       alevel (extendList env (map (fn (k,x) => (x, elevel)) binds)) e'
                   end
                 | Inject(_,l,e) => level e
                 | Match (_,e,cases) => 
                   let val elevel = alevel env e
                   in 
                       max (map (fn (MatchCase(_,l,x,e')) => alevel (extend env x elevel) e') cases) 
                   end
                 | Prom (_,e) => level e
                 | Der (_,e) => level e
                 | Force (_,e) => (level e) -- 1
                 | Fold (_,_,e) => level e
                 | Unfold (_,e) => level e
                 | Fun(_,x,t,e) => if sensitive env ee then  (alevel (extend env x 0) e) + 1 else 0
                 | App(_,e1,e2) => if level e2 = 0 then level e1 -- 1 else raise Impossible "Abstraction level >0 in application"
                 | Eq(_,e1,e2) => maxlevels [e1,e2]
                 | Infix(_,_,e1,e2) => maxlevels [e1,e2]
                 | If(_,e1,e2,e3) => maxlevels [e1,e2,e3]
                 | Rec(_,x,t,e) => raise Unimplemented 44
                 | Catchcont(_,x,t,e) => alevel (extend env x 0) e 
                 | Catchcopy(_,x,t,e) => alevel (extend env x 0) e 
                 | Cast(_,e,t) => level e
                 | Triv(_, linopt) => 0 
                 | Extend(_,linopt,e1,e2,e3) => if List.exists (sensitive env) [e1,e2,e3] then 
                                                    raise Impossible "Sensitive in extend" 
                                                else 0
                 | Constr (_,e) => if sensitive env e then raise Impossible "Sensitive in new" else 0
                 | Make (_,e1,e2) => if sensitive env e1 then raise Impossible "Sensitive in make" 
				     else if level e2 = 0 then 0
                                     else raise Impossible "Abstraction level >0 in make"
end
(* does some x : n >0 in env occur free in e? 
 * we just add binding variables with tag 0 to avoid counting these
 *)
and sensitive env e = 
    let val s = sensitive env 
        val ss = List.exists s
    in
    case e of Var(_,x) => (case lookup env x of SOME n => n > 0 | NONE => raise Impossible ("Sensitive: unknown variable "^x))
            | Const(_,ce) => false
	    | Product(_,es) => ss (map snd es)
	    | Merge(_,e1,e2)=> ss [e1,e2]
	    | Split(_,e1,binds,rest,e2)=> s e1 orelse sensitive (extendList (extend env rest 0) (map (fn (k,x) => (x, 0)) binds)) e2
	    | Inject(_,k,e) => s e
	    | Match(_,e,cases) => s e orelse List.exists (fn (MatchCase(_,l,x,e')) => sensitive (extend env x 0) e') cases
	    | Prom(_,e)=> s e
	    | Der(_,e)=> s e
	    | Force(_,e)=> s e
	    | Fold(_,_,e)=> s e
	    | Unfold(_,e)=> s e
	    | Fun(_,z,t,e)=> sensitive (extend env z 0) e
	    | App(_,e1,e2)=> ss [e1,e2]
	    | Eq(_,e1,e2)=> ss [e1,e2]
	    | Infix(_,_,e1,e2)=> ss [e1,e2]
	    | If(_,e1,e2,e3)=> ss [e1,e2,e3]
	    | Rec(_,z,t,e)=> sensitive (extend env z 0) e
	    | Catchcont(_,z,t,e)=> sensitive (extend env z 0) e
	    | Catchcopy(_,z,t,e)=> sensitive (extend env z 0) e
	    | Cast(_,e,t)=> s e
	    | Triv(_,l)=> false
	    | Extend(_,l,e1,e2,e3)=> ss [e1,e2,e3]
	    | Constr(_,e)=> s e
	    | Make(_,e1,e2)=> ss [e1,e2]
    end


local 
fun ccsum t1 t2 = SumTy([("result", t1), ("query", t2)])
fun mkp t t' = ProdTy([("value", t),("residue",t')],[])
fun catchXty t1 t2 t3 t4 maybebang = 
    let val more = maybebang (FunTy(BangTy(FunTy(t1,t2)), t4))
	val resume = maybebang (FunTy(BangTy t2,FunTy(BangTy(FunTy(t1,t2)),mkp t3 t4)))
    in
	  ccsum (ProdTy([("value", t3),("more", more)], []))
                (ProdTy([("arg", t1), ("resume", resume)], []))
    end
in
fun catchcontty t1 t2 t3 t4 = catchXty t1 t2 t3 t4 (fn x => x)
fun catchcopyty t1 t2 t3 t4 = catchXty t1 t2 t3 t4 BangTy
end



(* split binds labtys 
split lists of label-variable bindings k=x (binds) and label-type associations k=t (labtys) into 
 (1) those which match (2) leftover binds (3) leftover labtys
Confusingly return a tuple (3,1,2) 
*)
fun split binds labtys = split' binds labtys [] []
and split' ((k,x)::binds) labtys acc_vartys acc_unbound = 
    (case List.find (fn (k',_) => k=k') labtys 
      of SOME (_,t) => split' binds 
                              (List.filter (fn (k',_) => k<>k') labtys) 
                              ((x,t)::acc_vartys) 
                              acc_unbound
       | NONE => split' binds labtys acc_vartys (k::acc_unbound))
  | split' [] labtys acc_vartys acc_unbound = (labtys, rev acc_vartys, rev acc_unbound)


val unions = foldl VarSet.union VarSet.empty 
(* Union of pairwise intersections of each VarSet in a list *)
fun dups [] = VarSet.empty
  | dups (h::t) = VarSet.union(VarSet.intersection(h, unions t), dups t)

fun lincheck pos l = let val dupvars = dups l in
                 if VarSet.isEmpty dupvars then ()
         else raise Untypeable (pos, "linear variables reused: " ^ (concatWith ", " (VarSet.listItems dupvars)))
         end

(* Strangely convenient to pass the component k around here too *)
fun is_RW_ty (k, ProdTy(tys, [])) = 
    (case prod_match tys [("read", (fn (BangTy(FunTy(u1,sigma1))) => if u1 = unitty then SOME sigma1 else NONE | _ => NONE)),
			 ("write", (fn (BangTy(FunTy(sigma2,u2))) => if u2 = unitty then SOME sigma2 else NONE | _ => NONE))]
      of SOME [sigma1, sigma2] => if sigma1 =:= sigma2 then SOME (k, sigma1) else NONE
       | NONE => NONE
       | _ => raise Impossible "is_RW_ty - prod_match")
  | is_RW_ty _ = NONE

fun is_RW_all_ty (ProdTy(ts, [])) = 
      let val ts' = List.mapPartial is_RW_ty ts
      in if length ts = length ts' then SOME (ProdTy(ts', [])) else NONE
      end
  | is_RW_all_ty _ = NONE

fun is_meth_ty (k, BangTy(FunTy(t1,t2))) =
    (case is_RW_all_ty t1 of SOME ty => SOME ( (k, ty), (k, BangTy t2) )
			   | NONE => NONE)
  | is_meth_ty _ = NONE

(* is the given non-unit product type of the form t \natural t' ? 
   if so, return SOME (SOME t', t), or SOME (NONE, t) if t is empty so t' can't be determined *)
fun is_natural_ty (ProdTy([],[])) = SOME (NONE, ProdTy([],[]))
  | is_natural_ty (ProdTy(tys, [])) =
      let val res = List.mapPartial is_meth_ty tys
	  val (stys,res2) = ListPair.unzip res
          val sty = case stys of (_,sty)::_ => sty | _ => raise Impossible "No state types"
	  val same = List.all (fn (_,t) => t =:= sty) stys
      in if same andalso length tys = length res then SOME (SOME sty, ProdTy(res2,[]))
	 else NONE
      end
  | is_natural_ty _ = NONE

fun is_meth_ty_lin (k, BangTy(FunTy(ProdTy(t1,[]), ProdTy(t2,[])))) =
    (case prod_select t1 ["value", "state"] of 
       SOME [argty, sty] =>
	(case prod_select t2 ["value", "state"] of 
             SOME [BangTy resty, sty'] =>  
	     (* BangTy added to allow for backtracking on resulting method *)
	         if sty =:= sty' then SOME (sty, (k, BangTy(FunTy(argty,resty)))) else NONE
	   | NONE => NONE
	   | _ => raise Impossible "meth_ty_lin - prod match")
     | NONE => NONE
     | _ => raise Impossible "meth_ty_lin - prod match")
  | is_meth_ty_lin _ = NONE

(* is the given non-unit product type of the form t \natural_lin t' ? if so, return (t', t) *)
fun is_natural_ty_lin (ProdTy(tys, [])) =
    let val l =  List.mapPartial is_meth_ty_lin tys
	val (sigmas, tys') = ListPair.unzip l
    in if length l <> length tys then NONE
       else case sigmas of [] => SOME (NONE, ProdTy(tys',[]))
		    | (sigma::_) => if List.all (fn (sigma')=> sigma' =:= sigma) sigmas 
				    then SOME (SOME sigma, ProdTy(tys',[]))
				    else NONE
    end
  | is_natural_ty_lin _ = NONE

fun is_natural_ty_either linear t = if linear then is_natural_ty_lin t else is_natural_ty t

(* Suitable type for main body of class implementation? If so, return state type etc. *)
fun is_impl_ty lin pos (BangTy(FunTy(supty, BangTy(FunTy(selfty1, selfty2))))) = 
    (case (is_natural_ty_either lin supty, is_natural_ty_either lin selfty1, is_natural_ty_either lin selfty2) of 
        (SOME (sigma_sup, t_sup), SOME (SOME sigma1, t1), SOME (SOME sigma2, t2)) =>
        if sigma1 <:> sigma2 then 
            raise Untypeable(pos, "Mismatch between self and inferred field types in class implementation: "
                                  ^(tytostring sigma1)^" vs "^(tytostring sigma2))
        else if (case sigma_sup of SOME sigma0 => sigma1 <:> sigma0 | NONE => false) 
	     then raise Untypeable(pos, "Mismatch between self and super field types in class implementation: "^
                                                               (tytostring sigma1)^" vs "^(tytostring (Option.valOf sigma_sup)))
        else if t1 <:> t2 then raise Untypeable(pos, "Mismatch between self and inferred method types in class implementation: "^
                                                    (tytostring t1)^" vs "^(tytostring t2))
        else (sigma1, t_sup, t1)
      | (_,_,SOME(NONE,_)) => raise Untypeable (pos, "Body of class implementation needs at least one method: "^ (tytostring selfty2))
      | (_,SOME(NONE,_),_) => raise Untypeable (pos, "Self type in class implementation needs at least one method: "^ (tytostring selfty1))
      | (_,_,NONE) => raise Untypeable(pos, "Bad type for body of class implementation: "^ (tytostring selfty2))
      | (_,NONE,_) => raise Untypeable(pos, "Bad self type in class implementation: "^ (tytostring selfty1))
      | (NONE,_,_) => raise Untypeable(pos, "Bad super type in class implementation: "^ (tytostring supty)))
  | is_impl_ty _ pos t = raise Untypeable(pos, "Inappropriate type for class implementation function (should be t -> t' -> t'): \n"^(tytostring t))


(* Typechecking phase 1: Plain typechecking, regardless of tags.
   Produces a syntax tree with every subterm annotated with its type. *)

(* typecheck: envty -> envty -> errpos exp -> ty_safe -> ty * VarSet
   raise Untypeable on general typechecking error.
   Find type for exp in environment if possible with specific safety tag. 

  Note that environments are of the form (var * ty * ty_safe), while we
  always set the safety tag as Unsafevar - this is for compatibility with
  the top level environment. *)

fun typecheck tenv env (e: errpos exp) : ((errpos*ty) exp * VarSet.set) =
   let (* To derive a Safe judgement, first try a direct derivation, on failure try to derive an Unsafe judgement of ground type *)
       val status  = fn pos => fn t => status [] pos (resolvety pos tenv t)
       val statusE = fn pos => fn t => statusE  pos (resolvety pos tenv t)

       (* et - Expression Type (of a type-annotated expression as we return) *)
       val et = snd o exp_top : (errpos*ty) exp -> ty
       fun etres pos (e:(errpos*ty) exp) : ty = resolvety pos tenv (et e)

       fun is_ground_ty (pos:errpos) (t:ty) = is_ground (status pos (resolvety pos tenv t))
       fun is_reusable_ty (pos:errpos) (t:ty) = is_reusable (status pos (resolvety pos tenv t))

       fun p n = debugPrint ((Int.toString n) ^"\n")
       val typecheck'  = typecheck tenv
       val check = typecheck' env 

       val e_pos = exp_top e

       fun extend env x t = (x,t,Unsafevar)::env
       fun extendList env newenv = List.revAppend(map (fn (x,t) => (x,t,Unsafevar)) newenv,env) 
       fun lookup env x = 
           (case List.find (fn (y,_,_) => y=x) env of SOME (_,t,tag) => SOME (t)
                         | NONE => NONE)

       fun checkcatch pos cont x t e = 
         (case t of BangTy(FunTy(t1,t2)) => 
                    let val (e',vars) = typecheck' (extend env x t) e 
                    in case prodty_select (etres pos e') ["value", "residue"] of 
                           SOME [tl, tr] =>
                           if not (is_ground_ty pos tl) then 
                               raise Untypeable (pos,"Expected ground type, not "^(tytostring tl))
                           else if not (is_ground_ty pos t1) then 
                               raise Untypeable (pos,"Expected ground type, not "^(tytostring t1))
                                                (* check x not in env *)
                           else if cont then (Catchcont((pos,catchcontty t1 t2 tl tr),x,t,e'),vars)
                           else (* copy *)
                                    let val vars' = tryDelete(vars,x) in 
                                        if VarSet.isEmpty vars' then (Catchcopy((pos,catchcopyty t1 t2 tl tr), x,t,e'),vars')
                                        else raise Untypeable (pos,"Non-reusable variables found in catchcopy expression: "
                                                                   ^(concatWith ", " (VarSet.listItems vars'))) 
                                    end 
                         | _ => raise Untypeable (pos,"Catch body must have suitable product type, not "^(tytostring (et e')))
                    end
                  | _ => raise Untypeable (pos,"Catch variable "^x^" must have !(->) type, not " ^ (tytostring t)) 
         )

   in
     case e of Var(pos,x) => 
               (case lookup env x of SOME t => (Var((pos,t),x), if not (is_reusable_ty pos t) 
                                                          then VarSet.singleton x 
                                                          else VarSet.empty)
                                   | NONE => raise Untypeable(pos,"Unknown variable '"^x^"'"))
         | Const(pos,c) => let
                   val ty = case c of Bool b => BoolTy | Int i => IntTy
                   in (Const((pos,ty), c), VarSet.empty) end

         | Product(pos,l) => 
           let val (labels, exps) = ListPair.unzip l
               val dups = duplicates1 id labels
               val () = if dups = [] then () 
                        else raise Untypeable(pos,"Duplicate label(s) in product: " ^ (concatWith ", " dups))
               val (exps', vars) = ListPair.unzip (map (fn ex => check ex) exps)
               val () = lincheck pos vars
               val ty = ProdTy (ListPair.zip(labels, map et exps'), [])
               val l' = ListPair.zip(labels, exps')
           in (Product((pos,ty),l'), unions vars) end

         | Merge(pos, e1,e2) => 
           let val (e1', vars1) = check e1
               val (e2', vars2) = check e2
               val () = lincheck pos [vars1,vars2]
                                 (* check t1,t2 compatible *) (* remove left-duplicates *) (* ie implement oplus *)
               val t = case (etres pos e1',etres pos e2') 
                        of (ProdTy p1, ProdTy p2) => 
                           ProdTy(prod_oplus pos p1 p2)
                         | _ => raise Untypeable(pos,"Operands of merge must both be of product type")
           in (Merge((pos,t), e1', e2'), VarSet.union (vars1, vars2)) end

         | Split(pos, e1, binds, restvar, e2) => 
           let val (e1',vars) = check e1
               val (labtys,tvars) = case etres pos e1' of
                                ProdTy(tys, tvars) => (tys,tvars) 
                              | _ => raise Untypeable(pos,"Expected product type in split expression, not "
                                                          ^(tytostring (et e1')))
               val (rest, vartys, unbound) = split binds labtys 
               val () = if unbound = [] then () else 
                        raise Untypeable(pos,"Attempt to bind non-existent components in split: "
                                             ^ (concatWith ", " unbound))

               val (e2',vars') = typecheck' (extendList env ((restvar, ProdTy(rest,tvars))::vartys)) e2 

               val vars'' = VarSet.difference(vars', (* remove bound vars free in e2 *)
                                              VarSet.addList(VarSet.singleton restvar, 
                                                             map (fn (k,x)=>x) binds)) 
               val () = lincheck pos [vars, vars'']
           in (Split((pos,et e2'), e1', binds, restvar, e2'), VarSet.union(vars,vars'')) end

         | Inject(pos, k,e) => 
           let val (e', vars) = check e 
           in (Inject((pos,SumTy [(k,et e')]), k, e'), vars) end

         | Match(pos, e, cases) => 
           let val (e', vars) = check e 
               val labtys = case etres pos e' of 
                                SumTy(labtys) => labtys 
                              | _ => raise Untypeable(pos,"Non sum type in case expression: "^(tytostring (et e')))
               val labvarexps = map (fn (MatchCase(pos,l,x,e)) => (l,(x,e))) cases
               val (unbound, varexptys, extra) = split labvarexps labtys
               val () = if unbound <> [] then raise Untypeable(pos,"Missing cases: "^(concatWith ", " (map (fn (k,_)=>k) unbound)))
                        else if extra <> [] then raise Untypeable(pos,"Extra cases: "^(concatWith ", " extra))
                        else ()
               val (texps, vars') = ListPair.unzip (map (fn ((x,e),t) => let val (e', vars) = typecheck' (extend env x t) e 
                                                                         in (e', tryDelete(vars,x)) end )
                                                        varexptys)
               val () = app (fn casevars => lincheck pos [vars, casevars]) vars'
               val t' = case texps of e1::_ => et e1 | [] => raise Untypeable(pos,"Splitting empty sum!")
               val t'' = resolvety (0,0) tenv t'
               val cases' = map (fn ((l, (x, _)),e') => MatchCase((pos,t'),l,x,e')) 
                                (ListPair.zip (labvarexps,texps)) (* FIX: convoluted *)
				(* FIX: also clobbers pos *)
           in 
               if not (List.all (fn ee => etres (0,0) ee =:= t'') texps) then raise Untypeable(pos,"Cases must have matching types")
               else (Match((pos,t'),e',cases'), unions (vars::vars'))
           end

         | Prom(pos,v) => 
           let 
               val (e', vars) = check v
	       val () = check_value pos v
           in if not (VarSet.isEmpty vars) then
                  raise Untypeable(pos,"Non-reusable variables used in promoted expression: "
                                       ^ (concatWith ", " (VarSet.listItems vars)))
              else (Prom((pos,BangTy(et e')), e'),vars) 
           end

         | Der(pos,e) => 
           let val (e', vars) = check e
           in case etres pos e' of BangTy t' => (Der((pos,t'), e'), vars)
				 | t' => raise Untypeable(pos,"Cannot derelict non-! type "^(tytostring t'))
           end

         | Force(pos,e) => 
           let val (e', vars) = check e
           in case etres pos e' of BangTy (FunTy(ProdTy([],[]), t')) => (Force((pos,BangTy(t')), e'), vars)
				 | t' => raise Untypeable(pos,"Cannot force non-({}->t) type "^(tytostring t'))
           end

         | Fold(pos,t,e) => 
           let val (e', vars) = check e 
           in if (etres pos e') <: resolvety pos tenv (recbody pos tenv t) 
	      then (Fold((pos,t),t,e'), vars)
              else raise Untypeable (pos,"Argument to fold has incorrect type")
           end

         | Unfold(pos,e) => 
           let val (e', vars) = check e 
           in case etres pos e' of t => (Unfold ((pos, recbody pos tenv t), e'), vars)
           end

         | Fun(pos, x,t,e) => 
           let val _ = statusE pos t
               val (e', vars) = typecheck' (extend env x t) e
           in 
               (Fun((pos, FunTy(t, et e')), x, t, e'), tryDelete(vars, x)) 
           end

         | App(pos, e1,e2) => 
           let val (e1',vars1) = check e1 
               val (e2',vars2) = check e2 
               val () = lincheck pos [vars1,vars2]
           in 
              case etres pos e1' of FunTy(t,t') => 
                              if t =:= (etres pos e2') then (App((pos,t'), e1', e2'), VarSet.union(vars1,vars2))
                              else raise Untypeable(pos,"Type mismatch in application "^ (tytostring (et e1')) 
                                                        ^" vs "^(tytostring (et e2')) ^ (show_exps_nl [e1,e2]))
                            | _ => raise Untypeable(pos,"Non-function type in application: "^(tytostring (et e1')))
           end

         | Eq(pos,e1,e2) => 
           let val (e1',vars1) = check e1 
               val (e2',vars2) = check e2 
               val () = lincheck pos [vars1,vars2]
           in
               if (etres pos e1') <:> (etres pos e2') then 
                   raise Untypeable(pos,"Type mismatch in equality test: " ^ (tytostring (et e1')) ^" vs "^(tytostring (et e2')))
               else if not (is_ground_ty pos (et e1')) then
                   raise Untypeable(pos,"Equality testing not permitted for non-ground type: " ^ (tytostring (et e1')))
               else (Eq((pos,BoolTy), e1',e2'), VarSet.union(vars1,vars2))
           end

         | Infix(pos,i,e1,e2) =>
           let val (e1',vars1) = check e1
               val (e2',vars2) = check e2
               val () = lincheck pos [vars1,vars2]
	       val resultTy = (case i of Ssyn.IntLessThan => BoolTy
				       | _ => IntTy)
           in if (etres pos e1') <> IntTy orelse (etres pos e2') <> IntTy then
                   raise Untypeable(pos,"Infix arguments must both be integers: " ^ (tytostring (et e1')) ^"  "^(tytostring (et e2')))
              else (Infix((pos,resultTy), i,e1',e2'), VarSet.union(vars1,vars2))
           end

         | If(pos,e,e1,e2) => 
           let val (e',vars)  = check e 
               val (e1',vars1) = check e1
               val (e2',vars2) = check e2
               val () = (lincheck pos [vars,vars1]; lincheck pos [vars,vars2])
           in
               if (etres pos e') <> BoolTy then
                   raise Untypeable(pos,"Condition in 'if' expression not boolean type: "^(tytostring (et e')))
               else if (etres pos e1') <:> (etres pos e2') then
                   raise Untypeable(pos,"Type mismatch in branches of 'if' expression: " 
                                        ^ (tytostring (et e1')) ^" vs "^(tytostring (et e2')))
               else (If((pos,et e1'),e',e1',e2'), unions [vars,vars1,vars2])
           end

         | Rec(pos,x,t,e) => 
           let val _ = statusE pos t
	       val (e',vars) = typecheck' (extend env x t) e 
	       val vars = tryDelete(vars, x)
           in if not (groundless (etres pos e')) then
                  raise Untypeable(pos,"Variable "^x^" bound in rec not of groundless type: "^ (tytostring (et e')))
              else if (etres pos e') <:> resolvety pos tenv t then
                  raise Untypeable(pos,"Body of rec does not match type of variable "^x)
              else if not (VarSet.isEmpty vars) then
                  raise Untypeable(pos,"Variables used in body of rec expression must be reusable: "
                                        ^ (concatWith ", " (VarSet.listItems vars)))
              else (Rec((pos,et e'),x,t,e'),vars)
           end

         | Catchcont(pos,x,t,e) => (statusE pos t; checkcatch pos true x t e)
         | Catchcopy(pos,x,t,e) => (statusE pos t; checkcatch pos false x t e)

         | Cast(pos,e,t) => 
           let 
               val _ = statusE pos t
               val (e',vars) = check e 
           in if (etres pos e') </: (resolvety pos tenv t) then 
                  raise Untypeable(pos,"Cannot cast expression of type "
                                       ^(tytostring (et e'))^" to non-supertype "^(tytostring t))
              else (Cast((pos,t),e',t),vars)
           end

         | Triv(pos,lin) => 
           let val t = ClassImplTy(lin, ProdTy([],[]), ProdTy([],[]), ProdTy([],[]))
           in (Triv((pos,t), lin), VarSet.empty) end

         | Extend(pos,lin, e1,e2,e3) => 
           let val (e1', vars1) = check e1
               val (lo',pif,pim,tk) = 
                   case (etres pos e1') of ClassImplTy(lo,pif,pim, tk) =>(lo,pif,pim,tk)
                                         | _ => raise Untypeable(pos,"Superclass type not classimpl in extend: "^(tytostring (et e1')))
               val (e2', vars2) = typecheck' env e2
               val (sigma, body_pim, pim') = is_impl_ty lin pos (etres pos e2') 
               val () = if body_pim <:> pim 
			then raise Untypeable(pos, "Actual superclass method type does not match that of super")
			else if not (prod_inclusion pos  pim' pim) (* pim pim' *)
			then raise Untypeable(pos, "Class implementation type must contain superclass methods")
			else ()
					
               val pif'' = case sigma of ProdTy(p,_) => ProdTy(p,[])
                            (* OLD:          (case rev r of h::t => (ProdTy(p,t), h)
                                                      | _ => raise Untypeable(pos, "Field type missing row variable"))  *)
                                       | _ => raise Impossible "Field not product"
(* FIX *)
(*               val () = if prod_inclusion pos pif pif'' then ()
                        else raise Untypeable(pos, "Fields are not a superset of those in the superclass") *)
               val (e3', vars3) = check e3 
               val (tk'', maker_pif'') =
                   case (etres pos e3') of 
                       BangTy (FunTy (BangTy (FunTy (tk0, pif0)), BangTy (FunTy (tk1, pif1)))) =>
		           if tk0 =:= tk andalso pif0 =:= pif then (tk1, pif1)
			   else raise Untypeable (pos, "Type mismatch in constructor implementation")
		(* OLD: 
		       (case prod_match tys [("superarg", (fn tk'=>if tk=:=tk' then SOME tk'' else NONE)),
					     ("extender", (fn (FunTy(FunTy(u, pif'),pif'')) => 
                                                              if u = unitty andalso pif' =:= pif then SOME pif'' else NONE 
							    | tt => raise Untypeable(pos,"Bad extender type: "^(tytostring tt)) ))]
			 of SOME [t1,t2] => (t1,t2) 
			  | NONE =>  raise Untypeable(pos,"Type mismatch in extend")
                          | _ => raise Impossible "extend - prod_match")
                *)
                     | ty => raise Untypeable(pos,"Bad constructor type: "^(tytostring ty))

	       val () = if pif'' <:> maker_pif'' 
			then raise Untypeable(pos, "Field type in body and maker differ")

			else if not (VarSet.isEmpty vars2) 
			then raise Untypeable(pos,"Variables used in body of class implementation must be reusable: "
						  ^ (concatWith ", " (VarSet.listItems vars2)))
			else if not (VarSet.isEmpty vars3) 
			then raise Untypeable(pos,"Variables used in body of class constructor must be reusable: "
						  ^ (concatWith ", " (VarSet.listItems vars3)))
			else (check_value pos e2; check_value pos e3)

	       val lo = if lin = lo' then lo' else raise Untypeable(pos, "Cannot mix linear and non-linear classes")			   
                             
               fun addprod_replace (ProdTy(tys1,l1)) (ProdTy(tys2,l2)) = 
		   ProdTy (prod_oplus pos (tys1,l1) (tys2,l2))
                 | addprod_replace _ _ = raise Impossible "Expecting product type in \\oplus"

               val t = ClassImplTy(lo, pif'', addprod_replace pim pim', tk'')
		       (* FIX: check rho fresh *)
           in 
               statusE pos t;
               (Extend((pos,t),lin,e1',e2',e3'), unions [vars1,vars2,vars3])
           end
         | Constr (pos, e) => 
           let val (e',vars) = check e 
           in case etres pos e' of ClassImplTy(lin,t1,t2,t3) => 
                                   if not lin orelse is_ground_ty pos t3
				   then (Constr((pos,BangTy(FunTy(t3,t2))),e'), vars)
                                   else raise Untypeable(pos,"`new' disallowed for linear classes with non-ground constructor arguments - use `make'")
                          | t => raise Untypeable(pos,"Argument to new should have class implementation type, not "^ (tytostring t))
           end
         | Make (pos, e1, e2) =>
           let val (e1',vars1) = check e1
               val (e2',vars2) = check e2
           in case etres pos e1' of
               ClassImplTy(true,t1,t2,t3) =>
                    if t3 =:= etres pos e2' then (Make ((pos,t2), e1', e2'), VarSet.union(vars1,vars2))
                    else raise Untypeable (pos,"Second argument to make should have type " ^ (tytostring t3))
             | t => raise Untypeable(pos,"First argument to make should have linear classimpl type, not "^ (tytostring t))
           end
   end


(* Typechecking phase 2: New-safety checking. 
   Produces version of typed syntax tree with each subterm annotated 
   with its (best possible) newsafety status *)

val empty_env = []
fun extend env x tag = (x,tag)::env
fun lookup2 env x = 
  (case List.find (fn (y,_) => y=x) env of SOME (_,(tag1,tag2)) => SOME (tag1,tag2)
					 | NONE => NONE)
fun lookup env = Option.compose(fst, lookup2 env)
fun lookupnew env = Option.compose(snd, lookup2 env)


(* In the body of newcheck, read 'safe' as an abbreviation for 'newsafe' *)
fun newcheck tenv env (e: (errpos*ty) exp) : (errpos * ty * bool) exp =
let val check = newcheck tenv env 
    fun r newsafety = let val (pos,ty) = exp_top e in (pos, ty, newsafety) end
    fun safe (ee : (errpos*ty*bool) exp) = let val (_,_,safety) = exp_top ee in safety end
    val rsafe = r o safe
    fun allsafe es = List.all safe es
    val rallsafe = r o allsafe
    fun check2 (e1, e2) = (check e1, check e2)
    fun check3 (e1, e2, e3) = (check e1, check e2, check e3)

in
    case e of Var(_,x) => (case lookupnew env x of SOME Newsafevar => Var(r true, x)
							  | _ => Var(r false, x))
			    (* maybe ground (or !-free) variables should automatically be newsafe? *)
	      | Const(_,c) => Const(r true, c)
	      | Product(_,l) => let val l' = map (fn (x,y) => (x, check y)) l 
				in Product(rallsafe (map snd l'), l') end
	      | Merge(_,e1,e2) => let val (e1',e2') = check2 (e1,e2)
				  in Merge(rallsafe [e1',e2'], e1', e2') end
	      | Split(_,e1,binds,restvar,e2) => 
                     let val e1' = check e1
			 val stxn = (Unsafevar, if safe e1' then Newsafevar else Newunsafevar)
			 val e2' = newcheck tenv (extendList env ((restvar, stxn)::(map (fn (k,x) => (x,stxn)) binds))) e2 
		     in Split (r (safe e1' andalso safe e2'), e1',binds,restvar,e2') end 
		     (* doesn't seem to account for st2=halfnewsafe... done in safety check below? *)
              | Inject(_,k,e) => let val e' = check e in Inject (rsafe e', k, e') end
	      | Match(_,e,cases) => (* ugly AST transform case *)
                     let val e' = check e
			 val stxn = (Unsafevar, if safe e' then Newsafevar else Newunsafevar)
			 val cases' = map (fn MatchCase((pos,ty),l,x,e) => 
				                  let val e' = newcheck tenv (extend env x stxn) e
						  in MatchCase((pos,ty,safe e'),l,x, e') end) cases 
			 val result = List.all (fn MatchCase((_,_,s),_,_,_) => s) cases'
		     in Match(r result, e', cases') end (* e need not be safe *)
              | Prom(_,e) => let val e' = check e in Prom(rsafe e', e') end
              | Der(_,e) => let val e' = check e in Der(rsafe e', e') end
              | Force(_,e) => let val e' = check e in Force(rsafe e', e') end
              | Fold(_,t,e) => let val e' = check e in Fold(rsafe e', t, e') end
              | Unfold(_,e) => let val e' = check e in Unfold(rsafe e', e') end
              | Fun(_,x,t,e) => let val e' = newcheck tenv (extend env x (Unsafevar,Newsafevar)) e (* note no branching here *)
				in Fun(rsafe e', x, t, e') end
              | App(_,e1,e2) => let val (e1',e2') = check2 (e1,e2) in App (rallsafe [e1',e2'], e1',e2') end
              | Eq(_,e1,e2) => let val (e1',e2') = check2 (e1,e2) in Eq (rallsafe [e1',e2'], e1',e2') end
              | Infix(_,i,e1,e2) => let val (e1',e2') = check2 (e1,e2) in Infix (rallsafe [e1',e2'], i, e1',e2') end
              | If(_, e,e1,e2) => let val (e',e1',e2') = check3 (e,e1,e2) in If (rallsafe [e',e1',e2'], e', e1',e2') end
              | Rec(_,x,t,e) => let val e' = newcheck tenv (extend env x (Unsafevar,Newsafevar)) e (* note no branching here *)
				in Rec(rsafe e', x, t, e') end
              | Catchcont(_,x,t,e) => let val e' = newcheck tenv (extend env x (Unsafevar,Newsafevar)) e (* note no branching here *)
				in Catchcont(rsafe e', x, t, e') end
              | Catchcopy(_,x,t,e) => let val e' = newcheck tenv (extend env x (Unsafevar,Newsafevar)) e (* note no branching here *)
				in Catchcopy(rsafe e', x, t, e') end
              | Cast(_,e,t) => let val e' = check e in Cast(rsafe e', e', t) end
              | Triv(_,lin) => Triv(r true, lin)
              | Extend(_,lin,e1,e2,e3) => let val (e1',e2',e3') = check3 (e1,e2,e3) 
					  in Extend(rallsafe [e1',e2',e3'], lin, e1', e2',e3') end
              | Constr(_,e) => Constr(r false, check e) (* note not r (check e) *)
              | Make(_,e1,e2) => Make(r false, check e1, check e2)
end

fun halfnewcheck (ee: (errpos*ty*bool) exp) : bool  = 
    case ee of Constr(_,e) => (case exp_top e of
				  (_,ClassImplTy(true,_,_,_),_) => false
                                  (* looks fishy to me - JRL *)
				| (_,t,b) => b)
             | Make(_,e1,e2) => #3 (exp_top e1) andalso #3 (exp_top e2)
	     | Product(_,l) => List.all (halfnewcheck o snd) l
	     | _ => false


fun is_newsafe e = 
    let fun safety (_,_,safe) = safe 
    in safety (exp_top e) end

fun newsafetag e = if is_newsafe e then Newsafevar else Newunsafevar 

fun h_newsafetag e = if is_newsafe e then Newsafevar 
		     else if halfnewcheck e then Halfnewsafevar 
		     else Newunsafevar 


(* Typechecking phase 3: Argument safety checking. 
   Returns () on success, otherwise reports error. *)
			    
fun find_vars testf env vars =
  List.filter (fn x => case lookup2 env x of SOME tag => testf tag
					   | NONE => raise Impossible ("Variable "^x^" not found")) vars
fun find_labelled_vars testf env vars =
  List.filter (fn (x,lab) => case lookup2 env x of SOME tag => testf (tag,lab)
					   | NONE => raise Impossible ("Variable "^x^" not found")) vars
fun no_vars pos testf err env vars = 
  let val badvars = find_vars testf env vars
  in if badvars = [] then ()
     else raise Unsafe (pos, err ^ ": "^ (concatWith ", " badvars))
  end
fun no_labelled_vars pos testf err env vars = 
  let val badvars = find_labelled_vars testf env vars
  in if badvars = [] then ()
     else raise Unsafe (pos, err ^ ": "^ (concatWith ", " (map fst badvars)))
  end

fun safecheck tenv env (e: (errpos*ty*bool) exp) (tag: ty_safe) : unit = 
    let fun res (t : ty) = resolvety (0,0) tenv t
	fun et (e :(errpos * ty * bool) exp) : ty = res ((fn (_,z,_)=>z) (exp_top e))
	val (pos, ty, newsafe) = exp_top e

        fun drop_news exp = exp_map (fn (err,ty,newsafety) => (err,ty)) exp

        fun is_ground_ty t = is_ground (status [] pos t)
        fun unwantedArgOrMeta s = raise Unsafe (pos, s ^ " expression cannot be argsafe/metasafe") 
        fun checkArgOrMeta tag s = if tag = Argsafe orelse tag = Metasafe orelse tag = MetasafeLin then unwantedArgOrMeta s else ()
        fun unwantedMeta s = raise Unsafe (pos, s ^ " expression cannot be metasafe") 
        fun checkMeta tag s = if tag = Metasafe orelse tag = MetasafeLin then unwantedMeta s else ()
        val safecheck' = safecheck tenv
        fun check e tag = safecheck tenv env e tag

	fun safertag Newunsafevar = Weaknewsafevar
	  | safertag Weaknewsafevar = Weaknewsafevar
	  | safertag Halfnewsafevar = Newsafevar
	  | safertag Newsafevar = Newsafevar
        fun makesafe env = map (fn (x,(_,news)) => (x,(Safevar,safertag news))) env

       (* Some utilities, translating safety tags (and ST_ version) to variable tags too, and trying a bunch of alternative derivations *)
	fun st_trans ST_none = (Unsafevar, None)
	  | st_trans ST_safe = (Safevar, Safe)
	  | st_trans ST_argsafe = (Argsafevar, Argsafe)
	  | st_trans ST_writesafe = (Unsafevar, Writesafe)
	fun st_trans' tag = st_trans
	    (case tag of Ssyn.None => ST_none
		      | Ssyn.Safe => ST_safe
		      | Ssyn.Argsafe => ST_argsafe
		      | Ssyn.Writesafe => ST_writesafe 
		      | _ => raise Impossible "Metasafe found in st_trans'")
	fun try_alternatives f_totry (st1::st2::stl) =
              (f_totry (st_trans st1) handle ex => try_alternatives f_totry (st2::stl) handle ex' => raise ex) 
		  (* could 'merge' or note multiple e's raised*)
	  | try_alternatives f_totry [st1] = f_totry (st_trans st1)
	  | try_alternatives f_totry [] = raise Impossible "No alternatives"

        fun checkcatch pos cont x t e = 
           let val () = checkArgOrMeta tag "Catchcont"
	       val (stx, st)  = st_trans' tag
	       val stxn = (stx, Newunsafevar) (* note, not newsafetag e *)
	   in safecheck' (extend env x stxn) e st end

        fun destDot (Split (_,e,[(k,"__x")],"__rest",Var(_,x))) =
	    if x = "__x" then SOME (e,k) else NONE
	  | destDot _ = NONE

        val tag = if is_ground (status [] pos (et e)) andalso tag=Safe then None else tag

    in case e of 
           Var(_,x) => 
           (case lookup env x of 
		SOME tag' =>  
                if (case tag of Ssyn.None => true
                              | Ssyn.Safe =>  tag' = Safevar orelse tag' = RWvar
                              | Ssyn.Writesafe => tag' <> Writevar andalso tag' <> RWvar
                              | Ssyn.Argsafe => tag' = Argsafevar
                              | Ssyn.Metasafe => false 
			      | Ssyn.MetasafeLin => false)
		then () else 
                raise  Unsafe (pos, "Variable '"^x^"' has bad safety tag "
                                    ^(vtagtostring tag')
				    ^", wanted: "^(tagtostring tag))
              | NONE => raise Impossible ("Cannot not find variable: "^x))

         | Const(_,c) => checkArgOrMeta tag "Constant"

         | Product(_,l) => (checkMeta tag "Product";
                             app (fn (_,e) => check e tag) l)

         | Merge(_,e1,e2) => (checkMeta tag "Product";
                               check e1 tag; check e2 tag)

         | e0 as Split(_,e1,binds,restvar,e2) => 
           let fun do_split (stx,st) =
		  let val () = check e1 st
		      val stxn = (stx, h_newsafetag e1)
		  in safecheck' (extendList env ((restvar, stxn)::(map (fn (k,x) => (x,stxn)) binds))) e2 tag 
		  end
		  (* Hmm, if this applies ( split x : rwvar as ... in e end  ) then failure here =/> failure of other rule - 
                     writevar in ctx is a constraint! *)
		  (* "Not a writevar" should include rwvar, or the other rule should be excluded from application in that case *)
	       fun write_split () = 
                   (* First check for form x.k.read: this form is OK whether x is rwvar or not. *)
                   case (case destDot e0 of SOME (e0',"read") =>
					    (case destDot e0' of SOME (Var(_,x), _) => SOME x | _ => NONE)
					  | _ => NONE)
		    of SOME x => () 
		     | NONE => (* now try the general split rule *)
			  (check e1 Writesafe ; 
                           try_alternatives do_split [ST_safe, ST_none, ST_argsafe])

            (* Try all 3 cases for safety variable; first one hopefully common case, so give that error *)
            in (case tag of Ssyn.Safe => try_alternatives do_split [ST_safe, ST_none, ST_argsafe]
			  | Ssyn.Writesafe => write_split () 
			  | Ssyn.Argsafe => (try_alternatives do_split [ST_argsafe, ST_safe, ST_none]
                                        handle ex => 
                                           let val stxn = (Safevar, h_newsafetag e1) 
                                               val safeEnv = makesafe (extendList env ((restvar, stxn)::(map (fn (k,x) => (x,stxn)) binds)))
					   in  safecheck' safeEnv (newcheck tenv safeEnv (drop_news e2)) Argsafe
                                           end)
                                       (* for "let-argsafe rule", a late addition *)
			  | Ssyn.None => try_alternatives do_split [ST_none, ST_safe, ST_argsafe]
			  | _ => unwantedMeta "Split" (* try_alternatives do_split [ST_argsafe, ST_safe, ST_none] *)
	       )
            end

         | Inject(_,k,e) => (checkArgOrMeta tag "Product";
                              check e tag)

         | Match(_,e,cases) => 
           let fun do_match (stx,st) =  		  
               let val () = check e st
		   val stxn = (stx, newsafetag e)
	       in  app (fn MatchCase(pos,l,x,e) => (safecheck' (extend env x stxn) e tag)) cases end
            (* Try both cases for safety variable; first one hopefully common case, so give that error *)
            in case tag of Ssyn.Safe => try_alternatives do_match [ST_safe, ST_none]
                     | Ssyn.Writesafe => (check e Writesafe; try_alternatives do_match [ST_safe, ST_none])
                     | Ssyn.None => try_alternatives do_match [ST_none, ST_safe]
                     | _ => unwantedArgOrMeta "Case"
            end

         | Prom(_,e) => 
           let fun f_arg () = 
                   let val (x,_,e') = case e of Fun(_,x,t,e') => (x,t,e') 
					      | _ => raise Unsafe(pos, "fn body expected")
		       val env' = extend env x (RWvar, Newunsafevar)
		   in  debugPrint "Writesafe-checking body...\n" ; 
                       safecheck' env' e' Writesafe 
		   end
                                        
               fun f_meta () = 
                   let val (x,y,e') = case e of Fun(_,x,_,Prom(_,Fun(_,y,_,e'))) => (x,y,e')
                                             | _ => raise Unsafe(pos, "Metasafe function body not a double fn expression")
                       val env'  = extend env  x (Argsafevar,Newsafevar)
                       val env'' = extend env' y (Argsafevar,Newsafevar)
                   in safecheck' env'' e' Argsafe 
		   end

               fun f_metalin () = 
                   let val (x,y,e') = case e of Fun(_,x,_,Prom(_,Fun(_,y,_,e'))) => (x,y,e')
                                             | _ => raise Unsafe(pos, "Metasafe (linear) function body not a double fn expression")
                       val env'  = extend env  x (Unsafevar,Newunsafevar)
                       val env'' = extend env' y (Unsafevar,Newunsafevar)
                       val levenv = map (fn (x,_) => (x,0)) env
                       val levenv = extend (extend levenv x 1) y 1
                       val alevel = alevel levenv e'
                   in 
		      if alevel > 1 then raise Unsafe(pos, "'Abstraction level' of metasafe (linear) too high: "^ (Int.toString alevel))
		      else safecheck' env'' e' None 
		   end
           in case tag of
		  Ssyn.Argsafe => f_arg()
		| Ssyn.Metasafe => f_meta()
		| Ssyn.MetasafeLin => f_metalin()
		| _ => check e tag
	   end

         | Der(_,e) => (checkArgOrMeta tag "der"; check e tag)			      

         | Force(_,e) => (checkArgOrMeta tag "force"; check e tag)

         | Fold(_,_,e) =>   (checkArgOrMeta tag "fold"; check e tag)

         | Unfold(_,e) => (checkArgOrMeta tag "unfold"; check e tag)

         | Fun(_,x,_,e) => 
           let fun f (xtag,gtag) = safecheck' (extend env x (xtag, Newunsafevar)) e gtag 
					      (* FIX: check z, z' \notin body *)
               val all_weakly_safe = no_vars pos (fn (t,_)=> not (t=Safevar orelse t=RWvar)) 
					      "Non weakly safe variables in fn context" 
           in case tag of Ssyn.None => f (Unsafevar, None) 
                        | Ssyn.Safe => (f (Safevar, Safe);
                                   all_weakly_safe (extend env x (Safevar,Newunsafevar)) 
						   (get_vars e))
                        | Ssyn.Writesafe => f (Unsafevar, Writesafe)
			| _ => checkArgOrMeta tag "oncefn"
           end

         | App(_,e1,e2) =>
           let fun simple_app (tag1,tag2) =  (* Main application rule *)
                 (check e1 tag1; check e2 tag2)

	       val errs = "non (weakly) newsafe variables found in write context"

	       fun writesafe_app () = 
	           let val _ = (e1Debug := SOME e1 ; e2Debug := SOME e2) ;
                       val (trw,b) = 
                           case et e1 of FunTy(trw,_) => 
                                         (case is_RW_all_ty trw of SOME _ => (trw, true) 
                                                                 | NONE   => (trw, false))
				       | _ => raise Impossible ("Must be fun ty in writesafe app: "^(tytostring ty))
		   in if b then (debugPrint ("@Is RW ty: "^(tytostring trw)^"\n"); 
                                 case e1 of Der (_,e1') =>
                                     ((check e1' Argsafe ; check e2 Safe)
				      handle Unsafe(pos,s) => (debugPrint ("@52 Failed: "^s ^" trying @31\n"); 
                                                          simple_app (Writesafe,Writesafe)))
					  | _ => simple_app (Writesafe,Writesafe))
		      else case (case e1 of Der(_,e1') =>
				  (case destDot e1' of SOME (e1'',"write") => 
				    (case destDot e1'' of SOME (Var(_,x),_) => SOME x | _ => NONE)
						     | _ => NONE)
					  | _ => NONE)
			    of SOME x => (check_rw x ; check_newsafe x ; check_rw_ctxt x)
			     | NONE => simple_app (Writesafe, Writesafe)
		   end
	       and check_rw x =
		   (case lookup env x of 
			SOME RWvar => check e2 Safe
		      | SOME _ => raise Unsafe (pos, "Not a writevar in writesafe application: "^x)
		      | NONE => raise Impossible ("Unknown variable in writesafe application: "^x))
	       and check_newsafe x = 
		   if is_value e2 then 
                       (debugPrint "Checking if context is weakly newsafe.\n" ; e2Debug := SOME e2 ;
                        no_labelled_vars pos (fn ((_,tag2),(_,ty,_)) => 
						 (tag2=Newunsafevar orelse tag2=Halfnewsafevar) 
				                 andalso not (is_ground (status [] pos (resolvety pos tenv ty))))
                                         (* change to formal definition needed,
					    otherwise write arguments are unduly restricted. *)
					 errs env (get_labelled_vars e2))
                        (* Q: if this fails, should we then check if e2 is newsafe? *)
		   else if not (is_newsafe e2) then 
                       raise Unsafe (pos, "Argument to write is not newsafe")
                   else () 
	       and check_rw_ctxt x = 
		   no_vars pos (fn (t,_) => t=RWvar) 
			   "Argument to write contains read/write operations" env (get_vars e2)
           in
               case tag of Ssyn.None => simple_app (None, None)
                         | Ssyn.Safe => simple_app (Safe, Safe)
                         | Ssyn.Writesafe => writesafe_app ()
			 | Ssyn.Argsafe => (case e1 of
                                 (* concession to facilitate class impl/rep operations: *)
                                   Der (_, Var (_,"MAKE_ARGSAFE")) => check e2 None
                                 | _ => unwantedArgOrMeta "Application")
			 | _ => (case e1 of
                                   Der (_, Var (_,"MAKE_METASAFE")) => check e2 None
                                 | _ => unwantedArgOrMeta "Application")
           end             

         | Eq(_,e1,e2) => (checkArgOrMeta tag "Eq";
                           check e1 tag; check e2 tag)

         | Infix(_,_,e1,e2) => (checkArgOrMeta tag "Infix";
				check e1 tag; check e2 tag)

         | If(_, e,e1,e2) => (checkArgOrMeta tag "If";
                              check e tag; 
                              check e1 tag; check e2 tag)

         | Rec(_,x,t,e) => 
           let val () = checkArgOrMeta tag "Rec"
	       val (stx, st)  = st_trans' tag
	       val stxn = (stx, Newunsafevar)
	   in safecheck' (extend env x stxn) e st end

         | Catchcont(_,x,t,e) => checkcatch ty true x t e 

         | Catchcopy(_,x,t,e) => checkcatch ty false x t e 

         | Cast(_,e,t) => (checkMeta tag "Cast";
                           check e tag)

         | Triv(_,lin) => checkArgOrMeta tag "triv"

         | Extend(_,lin,e1,e2,e3) => 
           (checkArgOrMeta tag "Extend";
            check e1 tag; 
	    check e2 tag;
	    safecheck' (makesafe env) 
		       (newcheck tenv (makesafe env) (drop_news e2)) 
		       (if lin then MetasafeLin else Metasafe); 
	    check e3 tag)
	   (* linear metasafety should be different! *)

         | Constr(_,e) => (checkArgOrMeta tag "constr";
		           (* if tag = Writesafe then raise Unsafe(pos, "new expression cannot be writesafe") else *)
		           check e tag)

         | Make(_,e1,e2) => (checkArgOrMeta tag "make" ;
                             check e1 tag; check e2 tag)
    end
(* end of safecheck *)

end
end (* structure Typechecker *) ;
